;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - callback-edit.ss                                               #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Peter Ivanyi, 2007                                               #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

(module mreddesigner-callback-edit mzscheme
  (require (lib "class.ss")
           (lib "mred.ss" "mred")
           (file "graph-editor/graph-editor.ss")
           (file "mreddesigner-misc.ss")
           (file "mreddesigner-default-values.ss") ; for changed flag
  )
  
  
  ; ----------------------------------------------------------------------------  
  ; thin application layer to access the properties of the graph nodes
  
  (provide node-property-make)
  (define (node-property-make node)
    (send node set-data (make-hash-table 'equal))
  )
  
  (provide node-property-set!)
  (define (node-property-set! node prop val)
    (let
      ((table (send node get-data)))
      (hash-table-put! table prop val)
    )
  )
  
  (provide node-property-get)
  (define (node-property-get node prop)
    (let
      ((table (send node get-data)))
      (hash-table-get table prop #f)
    )
  )
  
  ; only these widgets can have output and callback functions
  (provide node-callbacks)
  (define node-callbacks '(button 
                           radio-box 
                           check-box 
                           list-box 
                           choice
                           text-field
                           slider))
  
  ; ----------------------------------------------------------------------------  
  
  ; this function turns a comma separated text into a list of texts
  (define (selection-list-process text)
    (define (trim-whitespace lst)
      (cond
        ((null? lst) '())
        ((char-whitespace? (car lst)) (trim-whitespace (cdr lst)))
        (else lst)
      )
    )
    (define (remove-empty lst)
      (cond
        ((null? lst) '())
        ((= (string-length (car lst)) 0) (remove-empty (cdr lst)))
        (else (cons (car lst) (remove-empty (cdr lst))))
      )
    )
    (remove-empty 
      (map 
        (lambda (str)
          (list->string (reverse (trim-whitespace (reverse (trim-whitespace (string->list str))))))
        )
        (text-split-with-empty text #\, #f)
      )
    )
  )
  
  ; ----------------------------------------------------------------------------
  
  (provide edit-action-window)
  (define (edit-action-window action-node src-id src-type dst-id dst-type)
    (letrec ((ok? #f)
             (window (new dialog%
                          (parent #f)
                          (spacing 0)
                          (width 350)
                          (alignment '(center top))
                          (label "Edit action")
                          (stretchable-width #f)
                          (stretchable-height #f)
                          ))
             ; source part
             (horizontal-panel-source (new horizontal-panel%
                                           (parent window)
                                           (min-height 25)
                                           (style '(border))
                                           (alignment '(left center))
                                           (stretchable-width #t)
                                           (stretchable-height #f)))
             (message-if (new message%
                              (parent horizontal-panel-source)
                              (label "if ")
                              (min-width 102)
                              (stretchable-width #f)
                              (stretchable-height #f)))
             (message-cond (new message%
                                (parent horizontal-panel-source)
                                (label "is pressed")
                                (stretchable-width #f)
                                (stretchable-height #f)))
             (choice-is-a (new choice%
                               (parent horizontal-panel-source)
                               (label #f)
                               (choices '("is" "is not"))
                               (horiz-margin 0)
                               (selection 0)
                               (stretchable-width #f)
                               (stretchable-height #f)))
             (text-field-cond (new text-field%
                                   (parent horizontal-panel-source)
                                   (min-width 100)
                                   (label #f)
                                   (init-value "")
                                   (horiz-margin 0)
                                   (vert-margin 2)
                                   (style '(single))
                                   (stretchable-width #f)
                                   (stretchable-height #f)))
             (choice-cond (new choice%
                               (parent horizontal-panel-source)
                               (label #f)
                               (choices '("true" "false"))
                               (horiz-margin 0)
                               (selection 0)
                               (min-width 100)
                               (min-height 0)
                               (stretchable-width #f)
                               (stretchable-height #f)))
             ;;; destination part
             (horizontal-panel-destination (new horizontal-panel%
                                                (parent window)
                                                (min-height 25)
                                                (alignment '(left center))
                                                (stretchable-height #f)
                                                (stretchable-width #t)
                                                (style '(border))))
             (message-then (new message%
                                (parent horizontal-panel-destination)
                                (label "then")
                                (min-width 100)
                                (stretchable-width #f)
                                (stretchable-height #f)))
             (choice-destination (new choice%
                                      (parent horizontal-panel-destination)
                                      (min-width 100)
                                      (label #f)
                                      (choices '())
                                      (selection 0)
                                      (stretchable-width #f)
                                      (stretchable-height #f)
                                      (callback 
                                       (lambda (c e)
                                         ; show the extra edit box if one of these are selected
                                         (if (member (send c get-string-selection) '("set-text"
                                                                                     "set-value"
                                                                                     "set-label"
                                                                                     "set-list" 
                                                                                     "set-selection"
                                                                                     ))
                                           (send horizontal-panel-destination 
                                                 change-children
                                                 (lambda (lst)
                                                   (list message-then
                                                         choice-destination
                                                         value-destination
                                                         message-destination-type
                                                         )
                                                   )
                                                 )
                                           (send horizontal-panel-destination 
                                                 change-children
                                                 (lambda (lst)
                                                   (list message-then
                                                         choice-destination
                                                         message-destination-type
                                                         )
                                                   )
                                                 )
                                           )
                                         )
                                      )))
             (value-destination (new text-field%
                                     (parent horizontal-panel-destination)
                                     (label "to  ")
                                     (style '(single deleted))
                                     (min-width 100)
                                     (stretchable-width #f)))
             (message-destination-type (new message%
                                            (parent horizontal-panel-destination)
                                            (min-width 100)
                                            (label "")
                                            (horiz-margin 4)
                                            (vert-margin 2)
                                            (stretchable-width #f)
                                            (stretchable-height #f)))
             ;;; button row
             ;;; ---------------------------------------------------
             (horizontal-panel-buttons (new horizontal-panel%
                                            (parent window)
                                            (vert-margin 6)
                                            (alignment '(center center))
                                            (stretchable-height #f)
                                            (stretchable-width #t)))
             (button-ok (new button%
                             (parent horizontal-panel-buttons)
                             (vert-margin 2)
                             (horiz-margin 2)
                             (min-width 60)
                             (label "OK")
                             (stretchable-width #f)
                             (stretchable-height #f)
                             (callback
                               (lambda (b e)
                                 (set! ok? #t)
                                 (send window show #f)))))
             (button-cancel (new button%
                                 (parent horizontal-panel-buttons)
                                 (vert-margin 2)
                                 (horiz-margin 2)
                                 (min-width 60)
                                 (label "Cancel")
                                 (stretchable-width #f)
                                 (stretchable-height #f)
                                 (callback
                                   (lambda (b e)
                                     (set! ok? #f)
                                     (send window show #f)))))
      )
      
      ; setup the above widgets
      ; =======================
      (let
        ; destination actions
        ((dst-options '()))

        ; setup source dependent row
        ; the row should contain which widgets
        (case src-type
          ((button)
           (send horizontal-panel-source
                 change-children
                 (lambda (lst)
                   (list message-if
                         message-cond)))
          )
          ((radio-box choice list-box text-field slider)
           (send horizontal-panel-source
                 change-children
                 (lambda (lst)
                   (list message-if
                         choice-is-a
                         text-field-cond)))
          )
          ((check-box)
           (send horizontal-panel-source
                 change-children
                 (lambda (lst)
                   (list message-if
                         choice-cond)))
          )
          (else
            (error "267: Unhandled source widget type: " src-type)
          )
        )
        ; set source label, e.g. "if button"
        (send message-if set-label (string-append "if " (symbol->string src-type)))
      
        ; setup destination dependent row
        ; determine what actions are possible, there are default actions
        ; and certain type of widgets allow special actions
        (set! dst-options '("enable" "disable" "show" "hide" "set-label"))
        (case dst-type
          ((check-box)
           (set! dst-options (append dst-options '("check" "uncheck" "toggle")))
          )
          ((radio-box)
           (set! dst-options (append dst-options '("set-selection")))
          )
          ((choice list-box)
           (set! dst-options (append dst-options '("set-list" "set-selection")))
          )
          ((text-field)
           (set! dst-options (append dst-options '("set-text")))
          )
          ((slider)
           (set! dst-options (append dst-options '("set-value")))
          )
          ; else the default
        )
        ; it is possible to append or hide the destination widget
        (set! dst-options (append dst-options '("append" "delete")))
        
        ; add the options to the choice widget
        (for-each (lambda (item) (send choice-destination append item)) dst-options)
      
        ; set destination type
        (send message-destination-type set-label (string-append "for " (symbol->string dst-type)))
        
        ; if there was an already specified action
        (let
          ((stored (node-property-get action-node 'callback-action)))
          (if stored
            (let*
              ((src      (list-ref stored 0)) ; the source condition
               (src-is-a (list-ref src 0))      ; is-a
               (src-val  (list-ref src 1))      ; the value
               (dst      (list-ref stored 1)) ; the destination action
               (action   (list-ref dst 0))      ; what to do
               (val      (list-ref dst 1))      ; with what value
               (idx      (- (length dst-options)
                            (length (member action dst-options))))
              )
              ; set source
              (case src-type
                ((button) (void))
                ((radio-box choice list-box text-field slider)
                 (if (equal? src-type 'slider)
                   (send text-field-cond set-value (number->string src-val))
                   (send text-field-cond set-value src-val)
                 )
                 (if src-is-a
                   (send choice-is-a set-string-selection "is")
                   (send choice-is-a set-string-selection "is not")
                 )
                )
                ((check-box)
                 (if (equal? src-val "true")
                   (send choice-cond set-selection 0)
                   (send choice-cond set-selection 1)
                 )
                )
                (else
                  (error "334: Unhandled source widget type: " src-type)
                )
              )
              ; set destination
              (send choice-destination set-selection idx)
              (cond
                ((member action '("enable" "disable" 
                                  "show" "hide" 
                                  "check" "uncheck" "toggle"
                                  "append" "delete"))
                 (void)
                )
                ((member action '("set-text" "set-value" "set-label"))
                 (send value-destination set-value val)
                )
                ((member action '("set-selection" "set-value"))
                 (send value-destination set-value (number->string val))
                )
                ((equal? action "set-list")
                 (let
                   ((str (car val)))
                   (for-each (lambda (x) (set! str (string-append str "," x))) (cdr val))
                   (send value-destination set-value str)
                 )
                )
                (else
                  (error "357: Unhandled action: " action)
                )
              )
              ; finally add the value widget to the destination row if required
              (if (member (send choice-destination get-string-selection)
                          '("set-text" "set-value" "set-label"
                            "set-list" "set-selection"))
                (send horizontal-panel-destination 
                      change-children
                      (lambda (lst)
                        (list message-then
                              choice-destination
                              value-destination
                              message-destination-type
                        )
                      )
                )
              )
            )
          )
        )
      )
      
      ; show the window
      ; ===============
      (send window show #t)
      
      ; if the OK button was pressed
      ; ============================
      (if ok?
        (let*
          ; determine the action to take on the destination
          ((action-type (send choice-destination get-string-selection))
           (action (cond
                     ((member action-type '("enable" "disable" 
                                            "show" "hide" 
                                            "check" "uncheck" "toggle"
                                            "append" "delete"))
                      (list action-type ""))
                     ((member action-type '("set-text" "set-label"))
                      (list action-type (send value-destination get-value))
                     )
                     ((member action-type '("set-selection" "set-value"))
                      (if (string->number (send value-destination get-value))
                        ; here we could check that the selected value is in range,
                        ; but the current list of choices may not be the one to which 
                        ; this selection corresponds
                        (let
                          ((idx (inexact->exact 
                                  (round (string->number (send value-destination get-value))))))
                          (list action-type idx)
                        )
                        (begin
                          (message-box "Warning"
                                       "Invalid number is entered, it is set to zero"
                                       #f '(ok caution))
                          (list action-type 0)
                        )
                      )
                     )
                     ((equal? action-type "set-list")
                      ; process the comma separated list
                      (list "set-list" 
                            (selection-list-process (send value-destination get-value)))
                     )
                     (else
                      (error "421: Unhandled action " action-type)
                     )
                   )
           )
           ; then generate a code depending on the source
           (callback (case src-type
                       ((button)
                        (list '(#f #f) action)
                       )
                       ((radio-box choice list-box text-field slider)
                        (let
                          ((is? (send choice-is-a get-string-selection))
                           (val (send text-field-cond get-value))
                          )
                          (list (list (if (string=? is? "is") #t #f)
                                      (if (equal? src-type 'slider)
                                        (string->number val) 
                                        val))
                                action)
                        )
                       )
                       ((check-box)
                        (list (list #t (send choice-cond get-string-selection)) action)
                       )
                       (else
                         (error "442: Unhandled source widget type: " src-type)
                       )))
          )
          ; store the action
          (node-property-set! action-node 'callback-action callback)
          (changed! #f)
        )
      )
    )
  )
  
) ; end of module


